// Vector.cpp
//
// ICS 65 Fall 2012
// Code Example
//
// This is an implementation of our Vector class.

#include "Vector.h"



// This notation is how you introduce definitions in a source file but hide
// them from other source files.  It puts the definitions into what's called
// the "unnamed namespace", which limits their availability only to this
// source file.
namespace
{
	// The default size of our Vectors when they're first created
	const unsigned int DEFAULT_VECTOR_SIZE = 10;


	// arrayCopy() is a utility function that copies "size" elements from
	// the "source" array into the "target" array.  It'll have use in a
	// few places throughout our implementation, so we'll pull it out and
	// make a function out of it.
	void arrayCopy(std::string* target, std::string* source, unsigned int size)
	{
		for (unsigned int i = 0; i < size; i++)
		{
			target[i] = source[i];
		}
	}
}



// Note that I've added commented-out lines that write output to "cout" that
// display a message when the constructor, copy constructor, destructor, and
// assignment operator are called.  Uncomment these and experiment with using
// the Vector class in various ways, so you can see which ones are called in
// which circumstances.



// The constructor need only allocate an array, set the size to 0, and set the
// capacity to whatever the length of the array is.  We can do all of that in
// the initializers, so there's no need for any code in the body (though I
// included console output, commented-out, for you to experiment with).

Vector::Vector()
	: items(new std::string[DEFAULT_VECTOR_SIZE]), sz(0), cap(DEFAULT_VECTOR_SIZE)
{
//	std::cout << "Vector::Vector()" << std::endl;
}


// The copy constructor initializes a new Vector to be a copy of an existing
// one, which necessitates a few things: (1) allocating a new array with the
// same capacity as the original, (2) copying the size and capacity into the
// new Vector, and (3) copying the individual elements from the old Vector
// into the new one.
//
// One thing to notice in this function (and in the assignment operator) is
// that we can legally access the private member variables of a Vector other
// than "this".  That's because access modifiers like "private" apply to a
// *class*, as opposed to *objects* of a class.  Code within a class can
// access *any* private member of *any* object of that class.  (And this is
// a sensible way to do business, because we trust the Vector class to know
// how to appropriately manage these member variables.)

Vector::Vector(const Vector& v)
	: items(new std::string[v.cap]), sz(v.sz), cap(v.cap)
{
//	std::cout << "Vector::Vector(const Vector&)" << std::endl;
	arrayCopy(items, v.items, sz);
}


// The destructor has only one job to do: deallocate the dynamically-allocated
// array of strings belonging to this Vector.  The member variables themselves
// (items, sz, and cap) will be destroyed automatically -- in the reverse of
// the order they're declared in the Vector class declaration -- just after
// the destructor ends, so we don't want to delete them here.

Vector::~Vector()
{
//	std::cout << "Vector::~Vector()" << std::endl;
	delete[] items;
}


// The assignment operator is trickier than the others in the Big Three,
// because of two bits of complexity:
//
// (1) It returns a reference to the Vector that's been assigned into.
//     Within any member function in Vector, there is a pointer called
//     "this" that points to the Vector the member function was called
//     on; in the case of the assignment operator, that's the Vector
//     that's been assigned into.  But we can't "return this" at the
//     end of the function, because "this" is a pointer and not a
//     reference.  To return a reference, we need to return an actual
//     Vector object, namely the one that "this" points to.  How do we
//     access the Vector object that "this" points to?  "*this".  So
//     we end the function with "return *this".
//
// (2) Self-assignment (assignment from one object into another) is
//     possible:
//
//         Vector v1;
//         ...
//         v1 = v1;
//
//     In this case, the assignment should simply have no effect.  But how
//     do we know that it's self-assignment?  We need to determine if the
//     object being assigned into has the same address as the object being
//     assigned from.  "this" is a pointer (the address!) of the object
//     being assigned into.  "v" is a reference (not the address, but
//     effectively the object!) to the object being assigned from.  However,
//     we can get the address of the object being assigned from by asking
//     for it: "&v".  So we'll check if "this" is different from "&v"; if
//     so, it's not self-assignment.

Vector& Vector::operator=(const Vector& v)
{
//	std::cout << "Vector& Vector::operator=(const Vector&)" << std::endl;

	if (this != &v)
	{
		std::string* newItems = new std::string[v.cap];
		arrayCopy(newItems, v.items, v.sz);

		sz = v.sz;
		cap = v.cap;

		delete[] items;
		items = newItems;
	}

	return *this;
}


// Notice that the bodies of these two functions appear identical.  What makes
// them different, though, is that they return different types: the first
// returns a reference to items[index] that will allow the value in that
// cell to be modified; the second returns a const reference to items[index],
// so it can be read but not modified.

std::string& Vector::at(unsigned int index)
{
	return items[index];
}


const std::string& Vector::at(unsigned int index) const
{
	return items[index];
}


// add() is straightforward, except when the array is full (i.e., sz == cap),
// in which case we need to allocate a new, larger array to take the place
// of the existing one.  Order is really important in that case; we don't
// want to delete the old array until *after* we've allocated the new one
// *and* copied all of the elements from the old one into the new one.

void Vector::add(const std::string& s)
{
	if (sz == cap)
	{
		int newCap = cap * 2 + 1;

		std::string* newItems = new std::string[newCap];
		arrayCopy(newItems, items, sz);

		cap = newCap;

		delete[] items;
		items = newItems;
	}

	items[sz] = s;
	sz++;
}


// size() and capacity() are the least interesting functions!  :)

unsigned int Vector::size() const
{
	return sz;
}


unsigned int Vector::capacity() const
{
	return cap;
}
