// Vector.h
//
// ICS 65 Fall 2012
// Code Example
//
// This is a class declaration for our Vector class, which implements an array-
// based list of strings.  Our Vector has three member variables:
//
// * A pointer to a dynamically-allocated array of strings
// * An unsigned integer that keeps track of the vector's size (i.e., how many
//   strings are currently stored in it)
// * Another unsigned integer that keeps track of the vector's capacity (i.e.,
//   how big the current array is)
//
// It provides modest amount of additional functionality; in a practical vector
// class, we'd need more than this, but these demonstrate the most interesting
// parts of what you might realistically build.
//
// * Read values given an index
// * Assign to values given an index and a new value
// * Add a new string to the end of the vector
// * Find out the vector's size
// * Find out the vector's capacity
//
// Our Vector is well-behaved, as well, which requires the Big Three, because
// we'll need to control what happens when a vector is destroyed, copied, or
// assigned.

#ifndef VECTOR_H
#define VECTOR_H

#include <string>



class Vector
{
public:
	// Constructing a vector makes it be empty and have a smallish initial
	// capacity.
	Vector();


	// This is the declaration for a "copy constructor".  C++ automaticaly
	// invokes this whenever a Vector object is constructed as a copy of
	// an existing one (e.g., when a Vector is passed by value, when a
	// Vector variable is declared as a copy of an existing one).
	//
	// If you don't write a copy constructor, one is generated automatically,
	// which simply does a member-by-member copy of the member variables, in
	// the order that they're declared.  In our case, that's problematic,
	// because we don't want to copy the pointer to the array of strings; we
	// want to copy the array itself.  So we need our own copy constructor.
	Vector(const Vector& v);


	// This is the declaration for a "destructor", recognizable because its
	// name is the name of the class preceded by a "~" character (and because
	// there is no return type).  C++ automatically invokes this just before
	// a Vector object dies.  After the destructor runs, the member variables
	// are themselves destroyed, in the reverse of the order they're declared
	// in the class.  But that's not enough in our case, because we don't just
	// want the pointer to the dynamically-allocated array to be destroyed;
	// we want the array itself to be destroyed.  The destructor will do that.
	~Vector();


	// This is the declaration for an "assignment operator".  In general, we
	// can redefine almost any operator in the C++ language by writing
	// functions whose name are "operator@" for any operator "@" (e.g.,
	// "operator=" for assignment, "operator+" for addition, "operator[]"
	// for indexing, etc.).
	//
	// Overloading operators requires understanding their underlying behavior
	// in more detail than you might be accustomed to.  In the case of
	// assignment, it's important to understand two things:
	//
	// (1) It's possible to do an assignment without modifying the object
	//     on the right-hand side, so we declare the parameter here to be
	//     const.
	// (2) Assignment is an expression that returns a value; specifically,
	//     assignment generally returns a reference to the object on the
	//     left-hand side (which we know will be an lvalue and, thus, can
	//     have a reference pointed to it).  While we could define our
	//     particular operator overload to return something different, it's
	//     best if we overload operators so they behave like their built-in
	//     counterparts, which follows the design "principle of least
	//     surprise".
	Vector& operator=(const Vector& v);


	// Unlike the version we wrote in lecture, the at() function is declared
	// here to return a reference to an actual string stored in the vector,
	// as opposed to returning a copy.  This has two effects:
	//
	// (1) It is a performance optimization, because it avoids copying the
	//     string in order to return it.
	// (2) It means we need different versions of the function for const
	//     Vectors than for non-const Vectors.  If the Vector is const, we
	//     have to return a const reference to one of its strings, because
	//     returning a non-const reference to the string would allow us to
	//     change it; if the vector is non-const, we can return a non-const
	//     reference to the string and allow it to be changed.  C++ will
	//     automatically call the appropriate overload of the at() function
	//     depending on whether we call it on a const or non-const Vector.
	//
	// Given that at() returns a reference, we can then use it as either an
	// lvalue or an rvalue:
	//
	//     cout << v.at(3) << endl;
	//     v.at(3) = "Boo";
	//
	// However, the latter of the two examples will only be permitted if the
	// Vector is non-const, because it requires the returned string reference
	// not to be const (or else we won't be allowed to assign to it).
	std::string& at(unsigned int index);
	const std::string& at(unsigned int index) const;


	// add() adds a new string to the end of the vector.  If the underlying
	// array is full, it is replaced by a new, larger array and the capacity
	// will grow as a result.
	void add(const std::string& s);


	// size() returns the size of the vector.
	unsigned int size() const;


	// capacity() returns the capacity of the vector.
	unsigned int capacity() const;


private:
	std::string* items;
	unsigned int sz;
	unsigned int cap;
};



#endif // VECTOR_H
